// This file is part of baretorrent and is released in the public
// domain. Hopefully, someone can make it better than I could.

#include "pch.h"
#include "dialogs.h"
#include "locale.h"
#include "utility.h"
#include "widgets.h"
#include "options_dialog.h"
#include "manager.h"

namespace baretorrent
{

enum ids
{
	browse_id = wxID_HIGHEST + 700,

	// reset button in the edit dialog
	reset_id,

	// advanced list
	advanced_list_id,

	// search box
	advanced_search_id,

	// log list
	log_list_id,

	// log level
	log_level_id,

	// clear log
	clear_log_id,

	// save log
	save_log_id,

	// extension list
	log_ext_list_id
};

class clipboard : boost::noncopyable
{
public:
	clipboard()
		: opened_(false)
	{
		if (!wxTheClipboard->Open())
		{
			log_error() << "couldn't open clipboard\n";
			return;
		}

		opened_ = true;
	}

	~clipboard()
	{
		if (opened_)
			wxTheClipboard->Close();
	}

	utf8_string get_text() const
	{
		if (!opened_)
			return utf8_string();

		if (wxTheClipboard->IsSupported(wxDF_TEXT))
		{
			wxTextDataObject d;
			wxTheClipboard->GetData(d);

			return utf8_string(d.GetText().ToUTF8());
		}

		return utf8_string();
	}

	void set_text(const utf8_string& s)
	{
		if (!opened_)
			return;

		wxTheClipboard->SetData(new wxTextDataObject(to_wxstring(s)));
	}

private:
	bool opened_;
};

add_dialog::add_dialog(wxWindow* parent) :
	wxDialog(
		parent, 0, loc_wx("torrent.add"),
		wxDefaultPosition, wxDefaultSize,
		wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER),
	text_(nullptr)
{
	text_ = new multiline_text(this, -1);

	clipboard cb;
	utf8_string s = cb.get_text();

	if (opt<bool>("bt.ui.add_from_clipboard"))
	{
		if (!s.empty() && plausible_torrent_string(s))
		{
			if (!boost::ends_with(s, "\n"))
				s += utf8_string("\n");
			
			text_->ChangeValue(to_wxstring(s));
			text_->SetSelection(
				static_cast<long>(s.size()),
				static_cast<long>(s.size()));
		}
	}

	wxFlexGridSizer* root = new wxFlexGridSizer(1, 5, 5);
	
	root->Add(new wxStaticText(this, 0, loc_wx("torrent.add.info")));
	root->Add(text_, 1, wxEXPAND);
	
	root->Add(new wxButton(
		this, browse_id, loc_wx("browse"),
		wxDefaultPosition, wxDefaultSize, wxBU_EXACTFIT));
	
	root->AddSpacer(5);
	
	root->Add(
		CreateSeparatedButtonSizer(wxOK | wxCANCEL),
		wxSizerFlags().Align(wxBOTTOM).Expand());

	root->AddGrowableCol(0);
	root->AddGrowableRow(1);

	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);
	box->Add(root, 1, wxEXPAND | wxALL, 5);

	text_->SetFocus();

	SetSizer(box);
	box->SetSizeHints(this);
	SetSize(500, 250);
	CentreOnParent();
}

std::vector<utf8_string> add_dialog::what() const
{
	return what_;
}

void add_dialog::on_browse(wxCommandEvent&)
{
	wxFileDialog dlg(
		this, loc_wx("browse.open_file"),
		wxEmptyString, wxEmptyString,
		loc_wx("wildcard.torrents"),
		wxFD_OPEN | wxFD_FILE_MUST_EXIST | wxFD_MULTIPLE);

	if (dlg.ShowModal() != wxID_OK)
		return;

	wxArrayString ss;
	dlg.GetPaths(ss);

	wxString t;

	foreach (const wxString& s, ss)
		t += s + to_wxstring(utf8_string("\n"));

	text_->AppendText(t);
}

void add_dialog::on_close(wxCommandEvent&)
{
	// saving urls
	what_ = explode(
		from_wxstring(text_->GetValue()), utf8_string("\n"));
	EndModal(wxID_OK);
}


wxDEFINE_EVENT(QUEUE_DIALOG_UPDATE_EVENT, wxCommandEvent);

queue_dialog::queue_dialog(wxWindow* parent, manager& m) :
	wxFrame(
		parent, 0, loc_wx("queue_dialog.title"),
		wxDefaultPosition, wxDefaultSize,
		wxRESIZE_BORDER | wxSYSTEM_MENU | wxCAPTION |
		wxCLIP_CHILDREN |wxFRAME_FLOAT_ON_PARENT),
	label_(nullptr), progress_(nullptr)
{
	wxPanel* root_panel = new wxPanel(this);

	wxPanel* panel = new wxPanel(root_panel);

	label_ = new wxStaticText(
		panel, -1, loc_wx("queue_dialog.start_label"));
	progress_ = new wxGauge(panel, -1, 0);

	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);
	box->Add(label_, 0, wxEXPAND);
	box->Add(progress_, 0, wxEXPAND);

	panel->SetSizer(box);

	wxBoxSizer* box2 = new wxBoxSizer(wxVERTICAL);
	box2->Add(panel, 1, wxALL | wxEXPAND, 5);

	root_panel->SetSizer(box2);

	wxBoxSizer* root = new wxBoxSizer(wxVERTICAL);
	root->Add(root_panel, 1, wxEXPAND);
	SetSizer(root);
	root->SetSizeHints(this);
	CentreOnParent();

	parent->Disable();

	Bind(
		QUEUE_DIALOG_UPDATE_EVENT,
		&queue_dialog::on_update_event, this);

	m.queue_progress.connect(boost::bind(
		&queue_dialog::on_update, this, _1, _2));
}

void queue_dialog::on_update_event(wxCommandEvent& e)
{
	std::auto_ptr<std::pair<std::size_t, std::size_t> > p(
		reinterpret_cast<std::pair<std::size_t, std::size_t>*>(
			e.GetClientData()));

	update(p->first, p->second);
}

void queue_dialog::on_update(std::size_t c, std::size_t t)
{
	wxCommandEvent* e = new wxCommandEvent(QUEUE_DIALOG_UPDATE_EVENT);
	e->SetEventObject(this);
	e->SetClientData(new std::pair<std::size_t, std::size_t>(c, t));

	QueueEvent(e);
}

void queue_dialog::update(std::size_t c, std::size_t t)
{
	if (c == t)
	{
		Destroy();
		GetParent()->Enable();
		return;
	}

	label_->SetLabel(to_wxstring(locf("queue_dialog.label") % c % t));

	progress_->SetRange(static_cast<int>(t));
	progress_->SetValue(static_cast<int>(c));
}


remove_dialog::remove_dialog(wxWindow* parent) :
	wxDialog(parent, 0, loc_wx("torrent.remove"),
		wxDefaultPosition, wxDefaultSize,
		wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER),
	del_(nullptr),
	keep_(true)
{
	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);

	box->AddSpacer(5);

	del_ = new wxCheckBox(
		this, -1, loc_wx("torrent.remove.delete_files"));

	del_->SetValue(!opt<bool>("bt.ui.remove_keeps_files"));

	box->Add(new wxStaticText(
		this, -1, loc_wx("torrent.remove.confirm")));
	box->AddSpacer(5);
	box->Add(del_);

	box->AddSpacer(5);

	box->AddStretchSpacer();
	box->Add(
		CreateSeparatedButtonSizer(wxOK | wxCANCEL),
		wxSizerFlags().Align(wxBOTTOM).Expand());

	wxBoxSizer* root = new wxBoxSizer(wxVERTICAL);
	root->Add(box, 1, wxEXPAND | wxALL, 5);

	SetSizerAndFit(root);
	CentreOnParent();
}

bool remove_dialog::keep_files() const
{
	return keep_;
}

void remove_dialog::on_close(wxCommandEvent&)
{
	keep_ = !del_->IsChecked();
	EndModal(wxID_OK);
}


bad_torrents_dialog::bad_torrents_dialog(
	wxWindow* parent, const std::vector<utf8_string>& v) :
		wxDialog(
			parent, -1, loc_wx("torrent.add.invalid_files.title"),
			wxDefaultPosition, wxDefaultSize,
			wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER)
{
	custom_list<int>* list =
		new custom_list<int>(this, -1, "", wxLC_NO_HEADER);
	
	list->add_column(utf8_string());

	foreach (const utf8_string& s, v)
		list->append(s);

	list->SetColumnWidth(0, wxLIST_AUTOSIZE);

	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);

	box->Add(new wxStaticText(
		this, 0, loc_wx("torrent.add.invalid_files")));
	box->Add(list, 1, wxEXPAND);
	box->AddSpacer(5);
	box->Add(
		CreateSeparatedButtonSizer(wxOK),
		wxSizerFlags().Align(wxBOTTOM).Expand());

	wxBoxSizer* root = new wxBoxSizer(wxVERTICAL);
	root->Add(box, 1, wxEXPAND | wxALL, 5);

	SetSizer(root);
	CenterOnParent();
}


add_trackers_dialog::add_trackers_dialog(wxWindow* parent) :
	wxDialog(
		parent, -1,
		loc_wx("connections.tracker.add.title"),
		wxDefaultPosition, wxDefaultSize,
		wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER),
	text_(nullptr)
{
	text_ = new multiline_text(this, -1);

	wxFlexGridSizer* root = new wxFlexGridSizer(1, 5, 5);
	
	root->Add(new wxStaticText(
		this, 0, loc_wx("connections.tracker.add.caption")));
	root->Add(text_, 1, wxEXPAND);
	
	root->AddSpacer(5);
	
	root->Add(
		CreateSeparatedButtonSizer(wxOK | wxCANCEL),
		wxSizerFlags().Align(wxBOTTOM).Expand());

	root->AddGrowableCol(0);
	root->AddGrowableRow(1);

	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);
	box->Add(root, 1, wxEXPAND | wxALL, 5);

	text_->SetFocus();

	SetSizer(box);
	SetSize(500, 250);
	CentreOnParent();
}

std::vector<utf8_string> add_trackers_dialog::trackers() const
{
	return what_;
}

void add_trackers_dialog::on_close(wxCommandEvent&)
{
	// saving urls
	what_ = explode(
		from_wxstring(text_->GetValue()), utf8_string("\n"));
	EndModal(wxID_OK);
}


about_dialog::about_dialog(wxWindow* parent)
	: wxDialog(parent, -1, loc_wx("about.title"),
		wxDefaultPosition, wxDefaultSize,
		wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER)
{
	wxSizer* s = CreateSeparatedButtonSizer(wxOK);

	wxStaticBitmap* image = new wxStaticBitmap(
		this, -1, platform::make_image(images::icon_128));
	
	wxStaticText* title = new wxStaticText(
		this, -1, to_wxstring(program_name()));
	
	wxStaticText* version = new wxStaticText(
		this, -1, to_wxstring(make_loc_version()));
	
	wxStaticText* target = new wxStaticText(
		this, -1,
		to_wxstring(locf("about.target") % platform::compile_target()));
	
	wxStaticText* pf = new wxStaticText(
		this, -1,
		to_wxstring(locf("about.platform") % platform::name()));

	wxHyperlinkCtrl* link = new wxHyperlinkCtrl(
		this, -1, to_wxstring(website_url()),
		to_wxstring(website_url()));
	
	// bigger program name label
	wxFont f = title->GetFont();
	f.SetPointSize(20);
	title->SetFont(f);
	
	wxBoxSizer* infos = new wxBoxSizer(wxVERTICAL);
	infos->Add(title);
	infos->Add(version);
	infos->Add(target);
	infos->Add(pf);
	infos->AddSpacer(20);
	infos->Add(link);
	
	wxGridBagSizer* grid = new wxGridBagSizer(5, 5);

	// from left to right: image, infos and versions
	grid->Add(image, wxGBPosition(0, 0), wxDefaultSpan, wxEXPAND);
	grid->Add(infos, wxGBPosition(0, 1), wxDefaultSpan, wxEXPAND);
	grid->Add(
		make_versions(), wxGBPosition(0, 2), wxDefaultSpan, wxEXPAND);

	grid->Add(s, wxGBPosition(1, 0), wxGBSpan(1, 3), wxEXPAND);

	grid->AddGrowableRow(0);
	grid->AddGrowableCol(2);

	wxBoxSizer* root = new wxBoxSizer(wxVERTICAL);
	root->Add(grid, 1, wxEXPAND | wxALL, 5);

	CentreOnParent();
	SetSizerAndFit(root);
}

wxSizer* about_dialog::make_versions()
{
	wxBoxSizer* box = new wxBoxSizer(wxHORIZONTAL);

	wxListCtrl* list = new wxListCtrl(
		this, -1, wxDefaultPosition, wxDefaultSize,
		wxLC_REPORT | wxLC_NO_HEADER);
	platform::set_list_control_style(list);

	// list with name and version columns
	list->InsertColumn(0, wxEmptyString);
	list->InsertColumn(1, wxEmptyString);

	std::vector<std::pair<utf8_string, utf8_string> > v =
		platform::versions();

	std::pair<utf8_string, utf8_string> p;
	foreach (p, v)
	{
		list->InsertItem(list->GetItemCount(), to_wxstring(p.first));

		list->SetItem(
			list->GetItemCount() - 1, 1, to_wxstring(p.second));
	}

	list->SetColumnWidth(0, wxLIST_AUTOSIZE);
	list->SetColumnWidth(1, wxLIST_AUTOSIZE);

	wxBitmap c = platform::make_image_unsafe(images::creeper);

	if (c.IsOk())
	{
		wxStaticBitmap* image = new wxStaticBitmap(this, -1, c);
		box->Add(image, 0, wxALIGN_BOTTOM);
	}

	box->Add(list, 1, wxEXPAND);

	// this makes sure the list is just large enough to see everything;
	// the +20 is for the padding around the list, which seems to work
	// fine on all platforms
	int w = list->GetColumnWidth(0) + list->GetColumnWidth(1) + 20;
	box->SetItemMinSize(list, w, -1);

	return box;
}


change_option_dialog::change_option_dialog(
	wxWindow* parent, const option& o) :
		wxDialog(
			parent, -1,
			loc_wx("options.advanced.change.caption"),
			wxDefaultPosition, wxDefaultSize,
			wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER),
		text_(nullptr),
		opt_(o)
{
	wxStaticText* label = new wxStaticText(
		this, -1, 
		loc_wx("options.advanced.change.message"));

	text_ = new wxTextCtrl(this, -1);

	wxButton* reset = new wxButton(
		this, reset_id, loc_wx("reset"));

	wxBoxSizer* tb = new wxBoxSizer(wxHORIZONTAL);
	tb->Add(text_, 1, wxEXPAND);
	tb->AddSpacer(5);
	tb->Add(reset);

	wxFlexGridSizer* fbox = new wxFlexGridSizer(1);

	fbox->Add(label, 0, wxEXPAND);
	fbox->Add(tb, 0, wxEXPAND);
	fbox->AddSpacer(5);
	fbox->Add(CreateButtonSizer(wxOK | wxCANCEL), 1, wxEXPAND);

	fbox->AddGrowableCol(0);

	wxBoxSizer* root = new wxBoxSizer(wxVERTICAL);
	root->Add(fbox, 1, wxEXPAND | wxALL, 5);

	SetSizerAndFit(root);
	SetSize(440, -1);

	text_->ChangeValue(to_wxstring(o.value()));

	text_->SelectAll();
	text_->SetFocus();
}

void change_option_dialog::on_reset(wxCommandEvent&)
{
	text_->ChangeValue(to_wxstring(opt_.default_value()));
}

void change_option_dialog::on_ok(wxCommandEvent&)
{
	opt_.set(utf8_string(text_->GetValue().ToUTF8()));
	EndModal(wxID_OK);
}

utf8_string change_option_dialog::value() const
{
	return opt_.value();
}


int option_list_name_sorter(option* left, option* right)
{
	return natural_compare(
		utf8_string(left->name()), utf8_string(right->name()));
}

int option_list_value_sorter(option* left, option* right)
{
	return natural_compare(left->value(), right->value());
}

option_list_widget::option_list_widget(
	wxWindow* parent, const std::string& guid) :
		wxPanel(parent), guid_(guid),
		list_(nullptr), description_(nullptr),
		search_(nullptr)
{
	splitter* split = new splitter(this, "bt.ui.advanced_splitter");

	list_ = new custom_list<option>(
		split, advanced_list_id, "bt.ui.advanced_list",
		wxLC_SINGLE_SEL);

	list_->add_column(
		loc("options.advanced.name"), &option_list_name_sorter);
	list_->add_column(
		loc("options.advanced.value"), &option_list_value_sorter);

	list_->selected.connect(boost::bind(
		&option_list_widget::on_selection, this, _1));

	list_->activated.connect(boost::bind(
		&option_list_widget::on_activated, this, _1));

	rebuild();

	description_ = new multiline_text(
		split, -1, wxEmptyString, wxDefaultPosition, wxDefaultSize,
		wxTE_READONLY);
	
	split->SplitHorizontally(list_, description_);
	split->SetSashGravity(1.0);

	wxStaticText* warning = new wxStaticText(
		this, -1, loc_wx("options.advanced.warning"),
		wxDefaultPosition, wxDefaultSize, wxST_NO_AUTORESIZE);

	wxFont font = warning->GetFont();
	font.SetWeight(wxFONTWEIGHT_BOLD);
	warning->SetFont(font);

	search_ = new wxTextCtrl(this, advanced_search_id);

	wxBoxSizer* search_box = new wxBoxSizer(wxHORIZONTAL);
	search_box->Add(new wxStaticText(
		this, -1, loc_wx("search")), 0, wxALIGN_CENTER_VERTICAL);
	search_box->AddSpacer(5);
	search_box->Add(search_, 1, wxEXPAND);

	wxFlexGridSizer* box = new wxFlexGridSizer(1, 5, 5);

	if (warning)
		box->Add(warning);

	box->Add(search_box, 1, wxEXPAND);
	box->Add(split, 1, wxEXPAND);

	box->AddGrowableCol(0);
	if (warning)
		box->AddGrowableRow(2);
	else
		box->AddGrowableRow(1);

	wxBoxSizer* root = new wxBoxSizer(wxVERTICAL);
	root->Add(box, 1, wxEXPAND | wxALL, 5);

	SetSizer(root);
}

option_list_widget::~option_list_widget()
{
	list_->delete_data();
}

void option_list_widget::rebuild(const utf8_string& search)
{
	list_->clear();
	
	option_list m;
		
	if (guid_.empty())
		m = get_options();
	else
		m = get_ext_options(guid_);

	std::pair<std::string, option> pair;

	foreach (pair, m)
	{
		if (pair.first == "bt.version")
			continue;

		if (!search.empty())
		{
			const std::string::size_type pn =
				pair.first.find(search.std_string());

			if (pn == std::string::npos)
			{
				const std::string::size_type pv =
					pair.second.value().std_string()
						.find(search.std_string());

				if (pv == std::string::npos)
					continue;
			}
		}
		
		int i = list_->append(utf8_string(pair.first));
		update_impl(i, pair.second);
	}
}

const option* option_list_widget::selected_option() const
{
	return list_->single_selection();
}

void option_list_widget::update_impl(int i, const option& o)
{
	list_->col(i, 1, o.value());

	if (list_->data(i) == nullptr)
		list_->data(i, new option(o));
	else
		list_->data(i)->set(o.value());

	wxFont f = list_->GetFont();
	if (o.value() != o.default_value())
		f.SetWeight(wxFONTWEIGHT_BOLD);

	list_->row_font(i, f);
}

void option_list_widget::on_selection(option* o)
{
	log_vv()
		<< "option list: selection changed to '" << o->name() << "'\n";

	utf8_string s;
	if (boost::starts_with(o->name(), "bt.session."))
	{
		s = utf8_string(
			"http://www.rasterbar.com/products/libtorrent/"
			"manual.html#session-settings");
	}
	else if (boost::starts_with(o->name(), "bt.dht."))
	{
		s = utf8_string(
			"http://www.rasterbar.com/products/libtorrent/manual.html"
			"#start-dht-stop-dht-set-dht-settings-dht-state-"
			"is-dht-running");
	}
	else if (!has_loc("option." + o->name()))
	{
		//bt_assert(false);
		log_error()
			<< "no description for option '" << o->name() << "'\n";
	}
	else
	{
		s = loc("option." + o->name());
	}

	description_->ChangeValue(to_wxstring(s));
}

void option_list_widget::on_search_changed(wxCommandEvent&)
{
	const utf8_string s(search_->GetValue().ToUTF8());
	rebuild(s);
}

void option_list_widget::on_activated(option* so)
{
	option op;
	if (new_[guid_].find(so->name()) != new_[guid_].end())
	{
		op = new_[guid_][so->name()];
	}
	else
	{
		if (guid_.empty())
			op = get_option(so->name());
		else
			op = get_ext_option(guid_, so->name());
	}
	

	change_option_dialog dlg(this, op);
	if (dlg.ShowModal() != wxID_OK)
		return;

	op.set(dlg.value());
	new_[guid_][op.name()] = op;
	
	fire(changed, guid_, op);
	update_impl(list_->single_selection_index(), op);
}


int log_date_sorter(
	const log_t::entry* left, const log_t::entry* right)
{
	if (left->time < right->time)
		return -1;
	else if (left->time > right->time)
		return 1;
	else
		return 0;
}

int log_message_sorter(
	const log_t::entry* left, const log_t::entry* right)
{
	return natural_compare(left->s, right->s);
}

int log_ext_name_sorter(
		const ext::instance* left, const ext::instance* right)
{
	return natural_compare(left->name(), right->name());
}


log_dialog::log_dialog(wxWindow*) :
	custom_frame("bt.ui.log_window"),
	list_(nullptr), ext_(nullptr), level_(log_t::normal_level)
{
	SetTitle(loc_wx("log.title"));
	
	splitter* split = new splitter(this, "bt.ui.log.split");

	list_ = new custom_list<log_t::entry>(
		split, log_list_id, "bt.ui.log_list");

	split->SetSashGravity(1.0);
	split->SplitVertically(list_, controls(split));

	list_->add_column(loc("log.time"), log_date_sorter);

	list_->add_column(loc("log.text"), log_message_sorter);

	new_entries_ = g_std_log.stop_capture_and_connect(
		boost::bind(&log_dialog::on_flushed, this, _1), &con_);

	timer_.start(
		boost::bind(&log_dialog::on_timer, this), 
		boost::posix_time::milliseconds(500));

	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);
	box->Add(split, 1, wxEXPAND);

	SetSizer(box);

	creating_ = false;

	list_->add_accelerator(wxGetStockAccelerator(wxID_COPY));
}

log_dialog::~log_dialog()
{
	con_.disconnect();
}

wxPanel* log_dialog::controls(wxWindow* parent)
{
	wxPanel* panel = new wxPanel(parent);

	wxGridBagSizer* grid = new wxGridBagSizer(5, 5);

	wxArrayString choices;
	choices.Add(loc_wx("log.level.error"));
	choices.Add(loc_wx("log.level.normal"));
	choices.Add(loc_wx("log.level.verbose"));
	choices.Add(loc_wx("log.level.very_verbose"));
	choices.Add(loc_wx("log.level.ext"));

	wxComboBox* lvs = new wxComboBox(
		panel, log_level_id, wxEmptyString,
		wxDefaultPosition, wxDefaultSize, choices,
		wxCB_READONLY);

	lvs->SetSelection(1);

	grid->Add(
		new wxStaticText(panel, -1, loc_wx("log.level")),
		wxGBPosition(0, 0), wxGBSpan(1, 1), wxALIGN_CENTER_VERTICAL);
	
	grid->Add(lvs, wxGBPosition(0, 1), wxGBSpan(1, 1), wxEXPAND);
	
	grid->Add(
		new wxStaticText(panel, -1, loc_wx("log.extensions")),
		wxGBPosition(1, 0), wxGBSpan(1, 2));


	ext_ = new custom_list<ext::instance>(
		panel, log_ext_list_id, "bt.ui.log_ext", 0);

	ext_->use_checkboxes();
	ext_->on_checked.connect(
		boost::bind(&log_dialog::on_ext_checked, this, _1, _2));

	ext_->add_column(loc("log.extensions.name"), log_ext_name_sorter);

	grid->Add(ext_, wxGBPosition(2, 0), wxGBSpan(1, 2), wxEXPAND);

	grid->Add(
		new multiline_text(panel, -1, to_wxstring(diag_info())),
		wxGBPosition(3, 0), wxGBSpan(1, 2), wxEXPAND);

	grid->Add(
		new wxButton(panel, clear_log_id, loc_wx("log.clear")),
		wxGBPosition(4, 0), wxGBSpan(1, 1));

	grid->Add(
		new wxButton(panel, save_log_id, loc_wx("log.save")),
		wxGBPosition(4, 1), wxGBSpan(1, 2));

	grid->AddGrowableCol(1);
	grid->AddGrowableRow(2);

	panel->SetSizer(grid);

	rebuild_ext();

	return panel;
}

void log_dialog::on_flushed(const log_t::entry& e)
{
	boost::mutex::scoped_lock lock(entries_mutex_);
	new_entries_.push_back(e);
}

void log_dialog::on_ext_checked(int, bool)
{
	rebuild(level_);
}

bool log_dialog::on_timer()
{
	log_t::entries v;

	{
		boost::mutex::scoped_lock lock(entries_mutex_);
		v = new_entries_;

		entries_.insert(entries_.end(), v.begin(), v.end());
		new_entries_.clear();

		if (log_t::max >= 0 &&
			static_cast<int>(entries_.size()) > log_t::max)
		{
			log_t::entries nv(
				entries_.begin() + entries_.size() - log_t::max,
				entries_.end());

			entries_.swap(nv);
		}
	}

	insert_items(v);
	return true;
}

void log_dialog::insert_items(const log_t::entries& v)
{
	bool scroll = false;
	int sel = -1;

	if (list_->GetItemCount() == 0)
	{
		scroll = true;
	}
	else if (list_->selection_size() == 1)
	{
		const int i = list_->single_selection_index();
		if (i == (list_->GetItemCount() - 1))
		{
			sel = i;
			scroll = true;
		}
	}

	freezer fz(list_);

	if (log_t::max >= 0)
	{
		int c = list_->GetItemCount();

		for (;;)
		{
			if (c == 0)
				break;
			
			if (static_cast<int>(c + v.size()) > log_t::max)
			{
				delete list_->data(0);
				list_->DeleteItem(0);
			}
			else
			{
				break;
			}

			--c;

			if (scroll && sel >= 0)
				--sel;
		}
	}
	
	ext::instance_list si;
	for (int i=0; i<ext_->GetItemCount(); ++i)
	{
		if (ext_->checked(i))
			si.push_back(ext::instances()[i]);
	}

	foreach (const log_t::entry& e, v)
	{
		if (level_ != log_t::quiet_level && e.lv > level_)
			continue;

		if (e.i)
		{
			if (std::find(si.begin(), si.end(), e.i) == si.end())
				continue;
		}
		else if (level_ == log_t::quiet_level)
		{
			continue;
		}
		
		const int i = list_->GetItemCount();

		list_->append(utf8_string(log_t::make_time(e.time)));
		list_->col(i, 1, e.s);
		list_->data(i, new log_t::entry(e));

		if (e.lv == log_t::error_level)
			list_->row_color(i, errored_text_color());
	}

	if (scroll && list_->GetItemCount() > 0)
	{
		if (sel >= 0)
			list_->unselect(sel);
		
		list_->select(list_->GetItemCount() - 1);
		list_->ensure_visible(list_->GetItemCount() - 1);
	}
}

void log_dialog::rebuild(log_t::levels lv)
{
	{
		boost::mutex::scoped_lock lock(entries_mutex_);

		log_t::entries v = entries_;
		v.insert(v.begin(), new_entries_.begin(), new_entries_.end());

		new_entries_.clear();
	}

	level_ = lv;
	list_->clear();
	insert_items(entries_);
}

void log_dialog::rebuild_ext()
{
	ext_->clear();

	ext::instance_list is = ext::instances();
	foreach (const ext::instance_ptr i, is)
	{
		const int item = ext_->append(i->name());
		ext_->data(item, i.get());
		ext_->checked(item, true);
	}
}

void log_dialog::on_clear_log(wxCommandEvent&)
{
	{
		boost::mutex::scoped_lock lock(entries_mutex_);
		entries_.clear();
		new_entries_.clear();
	}

	list_->clear();
}

void log_dialog::on_level(wxCommandEvent& e)
{
	log_level_indices i =
		static_cast<log_level_indices>(e.GetSelection());

	log_t::levels lv;
	switch(i)
	{
		case log_level_error:
			lv = log_t::error_level;
			break;

		case log_level_normal:
			lv = log_t::normal_level;
			break;

		case log_level_verbose:
			lv = log_t::verbose_level;
			break;

		case log_level_very_verbose:
			lv = log_t::very_verbose_level;
			break;

		case log_level_ext:
			lv = log_t::quiet_level;
			break;

		default:
			lv = log_t::normal_level;
			break;
	}

	rebuild(lv);
}

void log_dialog::on_close(wxCloseEvent&)
{
	Hide();
	opt("bt.ui.log", false);
}

void log_dialog::on_copy(wxCommandEvent&)
{
	std::ostringstream oss;

	const std::vector<const log_t::entry*> s = list_->selection();
	foreach (const log_t::entry* e, s)
		oss << *e << "\n";

	clipboard cb;
	cb.set_text(utf8_string(oss.str()));
}

void log_dialog::on_save_log(wxCommandEvent&)
{
	wxFileDialog dlg(
		this, loc_wx("log.save.title"), wxEmptyString, wxEmptyString,
		"*.*", wxFD_SAVE | wxFD_OVERWRITE_PROMPT);

	if (dlg.ShowModal() != wxID_OK)
		return;

	const utf8_path p(utf8_string(dlg.GetPath().ToUTF8().data()));
	if (p.empty())
		return;
	
	std::ofstream out(p.native().c_str(), std::ios::binary);
	if (!out)
		return;

	out
		<< to_simple_string(
			boost::posix_time::second_clock::local_time()) << "\n"
		<< diag_info()
		<< "\n\n"
		<< "extensions: ";

	ext::instance_list is = ext::instances();
	
	if (is.empty())
	{
		out << "none\n";
	}
	else
	{
		out << "\n";

		foreach (const ext::instance_ptr i, is)
			out << i->dump() << "\n\n";
	}

	out << "log:\n";

	const std::vector<const log_t::entry*> s = list_->items();
	foreach (const log_t::entry* e, s)
		out << *e << "\n";
}


power_dialog::power_dialog(wxWindow* parent, power_actions p)
	: wxDialog(parent, -1, wxEmptyString), label_(nullptr)
{
	std::string title;

	switch (p)
	{
		case shutdown_power:
		{
			title = "power.shutdown.title";
			message_ = locf("power.shutdown.warning");
			break;
		}

		case suspend_power:
		{
			title = "power.suspend.title";
			message_ = locf("power.suspend.warning");
			break;
		}

		case hibernate_power:
		{
			title = "power.hibernate.title";
			message_ = locf("power.hibernate.warning");
			break;
		}
			
		case no_power_action:	// fall-through
		default:
		{
			bt_assert(false);
			return;
		}
	}

	SetTitle(loc_wx(title));

	wxBoxSizer* box = new wxBoxSizer(wxVERTICAL);

	label_ = new wxStaticText(this, -1, wxEmptyString);
	box->Add(label_, 1, wxEXPAND);

	box->Add(
		CreateSeparatedButtonSizer(wxCANCEL),
		wxSizerFlags().Align(wxBOTTOM).Expand());

	wxBoxSizer* root = new wxBoxSizer(wxVERTICAL);
	root->Add(box, 1, wxEXPAND | wxALL, 5);

	SetSizerAndFit(root);
	CenterOnParent();
}

int power_dialog::ShowModal()
{
	end_ =
		boost::posix_time::second_clock::local_time() +
		boost::posix_time::seconds(opt<int>("bt.power_timeout"));

	set_label();
	Fit();

	timer_.start(
		boost::bind(&power_dialog::on_timer, this),
		boost::posix_time::milliseconds(100));

	return wxDialog::ShowModal();
}

bool power_dialog::on_timer()
{
	const boost::posix_time::ptime now =
		boost::posix_time::second_clock::local_time();

	if (now >= end_)
	{
		EndModal(wxID_OK);
		return false;
	}

	set_label();

	return true;
}

void power_dialog::set_label()
{
	const boost::posix_time::ptime now =
		boost::posix_time::second_clock::local_time();

	const boost::posix_time::time_duration d = end_ - now;

	const utf8_string m(
		(message_ % make_loc_duration(d.total_seconds())).str());

	wxString s = to_wxstring(m);

	if (label_->GetLabel() != s)
		label_->SetLabel(s);
}

BEGIN_EVENT_TABLE(log_dialog, wxFrame)
	EVT_BUTTON(clear_log_id, log_dialog::on_clear_log)
	EVT_BUTTON(save_log_id, log_dialog::on_save_log)
	EVT_COMBOBOX(log_level_id, log_dialog::on_level)
	EVT_CLOSE(log_dialog::on_close)
	EVT_MENU(wxID_COPY, log_dialog::on_copy)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(extension_options_dialog, wxDialog)
	EVT_BUTTON(wxID_OK, extension_options_dialog::on_ok)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(add_trackers_dialog, wxDialog)
	EVT_BUTTON(wxID_OK, add_trackers_dialog::on_close)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(add_dialog, wxDialog)
	EVT_BUTTON(browse_id, add_dialog::on_browse)
	EVT_BUTTON(wxID_OK, add_dialog::on_close)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(remove_dialog, wxDialog)
	EVT_BUTTON(wxID_OK, remove_dialog::on_close)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(change_option_dialog, wxDialog)
	EVT_BUTTON(wxID_OK, change_option_dialog::on_ok)
	EVT_BUTTON(reset_id, change_option_dialog::on_reset)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(option_list_widget, wxPanel)
	EVT_TEXT(advanced_search_id, option_list_widget::on_search_changed)
END_EVENT_TABLE()

}	// namespace baretorrent
